# curtain.py
#
# Sample Webots controller file for driving
# a "curtain" of actuated hanging chains.
#
# No copyright, 2020-2022, Garth Zeglin.
# This file is explicitly placed in the public domain.

from controller import Robot, Keyboard

import math


# Define the time step in milliseconds between
# controller updates.
EVENT_LOOP_DT = 20


# Request a proxy object representing the robot to control.
robot = Robot()
name = robot.getName()


# Query the number of devices.
# The curtain.proto model has one joint actuator and one sensor per chain.
num_devices = robot.getNumberOfDevices()
chains = num_devices // 2


# Enable computer keyboard input for user control.
keyboard = Keyboard()
keyboard.enable(EVENT_LOOP_DT)


# Fetch handles for the joint sensors.
# The names are generated by the curtain.proto scripting.
joints = [robot.getDevice("joint%d" % (jnum+1)) for jnum in range(chains)]


# Specify the sampling rate for the joint sensors.
for j in joints:
    j.enable(EVENT_LOOP_DT)


# Fetch handles for the position actuator at the top of each chain.
motors = [robot.getDevice("motor%d" % (jnum+1)) for jnum in range(chains)]
for m in motors:
    m.setPosition(0.0)


while robot.step(EVENT_LOOP_DT) != -1:
    # Read simulator clock time.
    t = robot.getTime()

    # Read any computer keyboard keypresses.
    # Returns -1 or an integer keycode while a key is held down.
    key = keyboard.getKey()

    if key != -1:
        # Convert the integer key number to a lowercase single-character string.
        letter = chr(key).lower()

        if letter == "q":
            # Enter a passive zero-torque mode.
            for m in motors:
                m.setTorque(0.0)
        elif letter == "e":
            # Drive to downward reference position.
            for m in motors:
                m.setPosition(0.0)
        elif letter == "w":
            # Drive all to front.
            for m in motors:
                m.setPosition(-0.5)
        elif letter == "s":
            # Drive all to back.
            for m in motors:
                m.setPosition(0.5)
        elif letter == "a":
            # Drive to alternating positions.
            for i, m in enumerate(motors):
                p = 0.5 if (i&1) == 0 else -0.5
                m.setPosition(p)
        elif letter == "d":
            # Drive to opposite alternating positions.
            for i, m in enumerate(motors):
                p = 0.5 if (i&1) == 1 else -0.5
                m.setPosition(p)
        elif letter == "x":
            # Generate a traveling wave (while 'w' is held down).
            for i, m in enumerate(motors):
                p = 0.5 * math.sin(1.5 * t + 0.75 * i)
                m.setPosition(p)
    else:
        if t % 2 < 0.25:
            if name.startswith("curtain_layer1"):
                for m in motors:
                    m.setPosition(0.5)
            else:
                for m in motors:
                    m.setPosition(-0.5)
        else:
            if name.startswith("curtain_layer1"):
                for m in motors:
                    m.setPosition(-0.5)
            else:
                for m in motors:
                    m.setPosition(0.5)
